from pydevmgr_vlt.devices.vltiodev.stat import VltIoDevStat as Stat
from pydevmgr_vlt.devices.vltiodev.ctrl  import VltIoDevCtrl as Ctrl

from pydevmgr_vlt.base import VltDevice
from pydevmgr_core import record_class, upload
from typing import Optional, Union, Iterable, Dict, List
from pydantic import BaseModel

Base = VltDevice


class VltioDevCtrlConfig(BaseModel):
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure (on top of CtrlConfig)
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    pass
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
class VltioDevConfig(Base.Config):
    CtrlConfig = VltioDevCtrlConfig
    
    Ctrl = Ctrl.Config
    Stat = Stat.Config
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure (redefine the ctrl_config)
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    type: str = "VltioDev"
    ctrl_config : CtrlConfig= CtrlConfig()
    
    ctrl: Ctrl = Ctrl()
    stat: Stat = Stat()
    
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



@record_class
class VltIoDev(Base):
    """ ELt Standard VltioDev device """
    Config = VltioDevConfig
    Ctrl = Ctrl
    Stat = Stat
    
    class Data(Base.Data):
        Ctrl = Ctrl.Data
        Stat = Stat.Data
        
        ctrl: Ctrl = Ctrl()
        stat: Stat = Stat()
    
    def init(self):
        """ init the iodev  """
        upload({
            self.ctrl.execute: True, 
            self.ctrl.command: self.ctrl.COMMAND.INITIALISE
            })
        return self.stat.initialised
    
    def set_do(self, flags: Union[List[bool], Dict[int,bool]]):
        """ set digital output flags 
        
        Args:
            flags (list, or dict): list of bool or a dictionary of digital output index (starting from 0) and flag pair
             
        Exemple::

            io.set_do( [False]*8 ) # set all to zero 
            io.set_do( [True, False] ) # set do_0 and do_1 to True and False respectively (others are unchaged)
            io.set_do( {3:True, 4:True} ) # set do_4 and do_4 to True (others are unchanged)
        """
        if not isinstance(flags, dict):
            it = enumerate(flags)
        else:
            it = flags.items()

        ctrl = self.ctrl 
        n_f = { getattr(ctrl, "do_{}".format(i)):f for i,f in it }
        
        n_f.update( {ctrl.execute:True, ctrl.command :self.ctrl.COMMAND.ACTIVATE} )
        upload(n_f)
    
    def set_ao(self, values: Union[List[bool], Dict[int,bool]]):
        """ set degital output values 
        
        Args:
            flags (list, or dict): list of float or a dictionary of analog output index (starting from 0) and value  pair
             
        Exemple::

            io.set_ao( [0.0]*8 ) # set all to zero 
            io.set_ao( [32, 64] ) # set do_0 and do_1 to 32 and 64 respectively (others are unchaged)
            io.set_ao( {3:128, 4:128} ) # set do_4 and do_4 to 128 (others are unchanged)
        """

        if not isinstance(values, dict):
            it = enumerate(values)
        else:
            it = values.items()

        ctrl = self.ctrl 
        n_f = { getattr(ctrl, "do_{}".format(i)):f for i,f in it }
        
        n_f.update( {ctrl.execute:True, ctrl.command :self.ctrl.COMMAND.ACTIVATE} )
        upload(n_f)
    

    def get_do_node(self, num: Union[int,Iterable]):
        if hasattr(num, "__iter__"):
            return [ getattr( self.ctrl, f'do_{n}') for n in num]
        else:
            return getattr( self.ctrl, f'do_{num}')

    def get_ao_node(self, num: Union[int,Iterable]):
        if hasattr(num, "__iter__"):
            return [ getattr( self.ctrl, f'ao_{n}') for n in num]
        else:
            return getattr( self.ctrl, f'ao_{num}')

    def get_ai_node(self, num: Union[int,Iterable]):
        if hasattr(num, "__iter__"):
            return [ getattr(self.stat, f'ai_{n}') for n in num]
        else:
            return getattr(self.stat, f'ai_{num}')
     
    def get_di_node(self, num: Union[int,Iterable]):
        if hasattr(num, "__iter__"):
            return [ getattr( self.stat, f'di_{n}') for n in num]
        else:
            return getattr( self.stat, f'di_{num}')

if __name__ == "__main__":
    VltIoDev()
    print("OK")



from pydevmgr_core import  NodeAlias1, Defaults, NodeVar, record_class, BaseParser
from pydevmgr_vlt.base import VltDevice
from pydevmgr_vlt.devices._tools import _inc
from pydevmgr_ua import Int32
from enum import Enum
from pydantic import create_model
Base = VltDevice.Ctrl

N = Base.Node # Base Node
NC = N.Config
ND = Defaults[NC] # this typing var says that it is a Node object holding default values 
NV = NodeVar # used in Data 

class COMMAND(int, Enum):
    NONE = 0
    INITIALISE = 1
    ACTIVATE = 2


N_AO, N_DO, N_NO, N_TO  = [8]*4



@record_class
class IoDevCommand(BaseParser):
    class Config(BaseParser.Config):
        type = "IoDevCommand"
    @staticmethod
    def parse(value, config):
        if isinstance(value, str):
            value =  getattr(COMMAND, value)
        return Int32(value)



# some dinamicaly created nodes
io_nodes = {}
for i in range(N_DO):
    io_nodes[f'do_{i}'] = (ND, NC(suffix= f'ctrl.arr_DO[{i}].bValue'))
for i in range(N_AO):
    io_nodes[f'ao_{i}'] = (ND, NC(suffix= f'ctrl.arr_AO[{i}].lrValue'))
for i in range(N_NO):
    io_nodes[f'no_{i}'] = (ND, NC(suffix= f'ctrl.arr_NO[{i}].nValue'))
for i in range(N_TO):
    io_nodes[f'to_{i}'] = (ND, NC(suffix= f'ctrl.arr_TO[{i}].sValue'))



class VltIoDevCtrl(Base):
    COMMAND = COMMAND
    class Config(create_model("Config",  __base__ = Base.Config, **io_nodes)):
        execute: ND = NC(suffix= 'ctrl.bExecute', parser= 'bool' )
        command: ND = NC(suffix= 'ctrl.nCommand', parser= 'IoDevCommand' )
        
    class Data(Base.Data):
        pass          

if __name__ == "__main__":
    VltIoDevCtrl()
    print("OK")




from pydevmgr_vlt.base import VltDevice
from pydevmgr_core import Defaults, NodeVar,   record_class, BaseParser
from pydevmgr_ua import Int32
from enum import Enum 
from typing import Optional 
from pydevmgr_vlt.devices.vltmotor.positions import PositionsConfig
from pydevmgr_vlt.devices.vltmotor.init_seq import INITSEQ

Base = VltDevice.Cfg

N = Base.Node # Base Node
NC = N.Config
ND = Defaults[NC] # this typing var says that it is a Node object holding default values 
NV = NodeVar # used in Data 



class AXIS_TYPE(int, Enum):
    LINEAR = 1
    CIRCULAR =2
    CIRCULAR_OPTIMISED = 3



def axis_type(axis_type):
    """ return always a axis_type int number from a number or a string
    
    Raise a ValueError if the input string does not match axis type
    Example:
        axis_type('LINEAR') == 1
        axis_type(1) == 1
    """
    if isinstance(axis_type, str):
        try:
            axis_type = getattr(AXIS_TYPE, axis_type) 
        except AttributeError:
            raise ValueError(f'Unknown AXIS type {axis_type!r}')
    return Int32(axis_type)

# a parser class for axis type
@record_class
class AxisType(BaseParser):
    class Config(BaseParser.Config):
        type: str = "AxisType"
    @staticmethod
    def parse(value, config):
        return axis_type(value)   




class VltMotorCfg(Base):
    AXIS_TYPE = AXIS_TYPE
    INITSEQ = INITSEQ
    class Config(Base.Config):
        scale_factor:       ND  =  NC(  suffix=  'cfg.lrScaleFactor'                )
        accel:              ND  =  NC(  suffix=  'cfg.lrAccel'                      )
        decel:              ND  =  NC(  suffix=  'cfg.lrDecel'                      )
        jerk:               ND  =  NC(  suffix=  'cfg.lrJerk'                       )
        backlash:           ND  =  NC(  suffix=  'cfg.lrBacklash'                   )
        velocity:           ND  =  NC(  suffix=  'cfg.lrDefaultVelocity'            )
        max_pos:            ND  =  NC(  suffix=  'cfg.lrMaxPosition'                )
        min_pos:            ND  =  NC(  suffix=  'cfg.lrMinPosition'                )
        tolerence:          ND  =  NC(  suffix=  'cfg.lrTolerance'                  )
        tolerence_enc:      ND  =  NC(  suffix=  'cfg.lrToleranceEnc'               )
        axis_type:          ND  =  NC(  suffix=  'cfg.nTypeAxis',                   parser=  'AxisType'  )
        tout_init:          ND  =  NC(  suffix=  'cfg.tTimeoutInit',                parser=  'UaInt32'   )
        tout_move:          ND  =  NC(  suffix=  'cfg.tTimeoutMove',                parser=  'UaInt32'   )
        tout_switch:        ND  =  NC(  suffix=  'cfg.tTimeoutSwitch',              parser=  'UaInt32'   )
        brake:              ND  =  NC(  suffix=  'cfg.bUseBrake'                    )
        low_brake:          ND  =  NC(  suffix=  'cfg.bActiveLowBrake'              )
        active_low_lstop:   ND  =  NC(  suffix=  'cfg.bArrActiveLow[0].bActiveLow'  )
        active_low_lhw:     ND  =  NC(  suffix=  'cfg.bArrActiveLow[1].bActiveLow'  )
        active_low_ref:     ND  =  NC(  suffix=  'cfg.bArrActiveLow[2].bActiveLow'  )
        active_low_index:   ND  =  NC(  suffix=  'cfg.bArrActiveLow[3].bActiveLow'  )
        active_low_uhw:     ND  =  NC(  suffix=  'cfg.bArrActiveLow[4].bActiveLow'  )
        active_low_ustop:   ND  =  NC(  suffix=  'cfg.bArrActiveLow[5].bActiveLow'  )
        init_seq1_action:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[1].nAction',    parser=  'UaInt32'   )
        init_seq1_value1:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[1].lrValue1'    )
        init_seq1_value2:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[1].lrValue2'    )
        init_seq2_action:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[2].nAction',    parser=  'UaInt32'   )
        init_seq2_value1:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[2].lrValue1'    )
        init_seq2_value2:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[2].lrValue2'    )
        init_seq3_action:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[3].nAction',    parser=  'UaInt32'   )
        init_seq3_value1:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[3].lrValue1'    )
        init_seq3_value2:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[3].lrValue2'    )
        init_seq4_action:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[4].nAction',    parser=  'UaInt32'   )
        init_seq4_value1:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[4].lrValue1'    )
        init_seq4_value2:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[4].lrValue2'    )
        init_seq5_action:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[5].nAction',    parser=  'UaInt32'   )
        init_seq5_value1:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[5].lrValue1'    )
        init_seq5_value2:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[5].lrValue2'    )
        init_seq6_action:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[6].nAction',    parser=  'UaInt32'   )
        init_seq6_value1:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[6].lrValue1'    )
        init_seq6_value2:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[6].lrValue2'    )
        init_seq7_action:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[7].nAction',    parser=  'UaInt32'   )
        init_seq7_value1:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[7].lrValue1'    )
        init_seq7_value2:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[7].lrValue2'    )
        init_seq8_action:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[8].nAction',    parser=  'UaInt32'   )
        init_seq8_value1:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[8].lrValue1'    )
        init_seq8_value2:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[8].lrValue2'    )
        init_seq9_action:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[9].nAction',    parser=  'UaInt32'   )
        init_seq9_value1:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[9].lrValue1'    )
        init_seq9_value2:   ND  =  NC(  suffix=  'cfg.strArrInitSeq[9].lrValue2'    )
        init_seq10_action:  ND  =  NC(  suffix=  'cfg.strArrInitSeq[10].nAction',   parser=  'UaInt32'   )
        init_seq10_value1:  ND  =  NC(  suffix=  'cfg.strArrInitSeq[10].lrValue1'   )
        init_seq10_value2:  ND  =  NC(  suffix=  'cfg.strArrInitSeq[10].lrValue2'   )

    class Data(Base.Data):

        scale_factor:       NV[float]  =  1.0
        accel:              NV[float]  =  30.0
        decel:              NV[float]  =  30.0
        jerk:               NV[float]  =  100.0
        brake:              NV[bool]   =  False
        backlash:           NV[float]  =  0.0
        axis_type:          NV[int]    =  0
        velocity:           NV[float]  =  0.0
        max_pos:            NV[float]  =  0.0
        min_pos:            NV[float]  =  0.0
        tolerence:          NV[float]  =  1.0
        tolerence_enc:      NV[int]    =  100
        tout_init:          NV[int]    =  0
        tout_move:          NV[int]    =  0
        tout_switch:        NV[int]    =  0
        low_brake:          NV[bool]   =  False
        active_low_lstop:   NV[bool]   =  False
        active_low_lhw:     NV[bool]   =  False
        active_low_ref:     NV[bool]   =  False
        active_low_index:   NV[bool]   =  False
        active_low_uhw:     NV[bool]   =  False
        active_low_ustop:   NV[bool]   =  False
        init_seq1_action:   NV[int]    =  0
        init_seq1_value1:   NV[float]  =  0.0
        init_seq1_value2:   NV[float]  =  0.0
        init_seq2_action:   NV[int]    =  0
        init_seq2_value1:   NV[float]  =  0.0
        init_seq2_value2:   NV[float]  =  0.0
        init_seq3_action:   NV[int]    =  0
        init_seq3_value1:   NV[float]  =  0.0
        init_seq3_value2:   NV[float]  =  0.0
        init_seq4_action:   NV[int]    =  0
        init_seq4_value1:   NV[float]  =  0.0
        init_seq4_value2:   NV[float]  =  0.0
        init_seq5_action:   NV[int]    =  0
        init_seq5_value1:   NV[float]  =  0.0
        init_seq5_value2:   NV[float]  =  0.0
        init_seq6_action:   NV[int]    =  0
        init_seq6_value1:   NV[float]  =  0.0
        init_seq6_value2:   NV[float]  =  0.0
        init_seq7_action:   NV[int]    =  0
        init_seq7_value1:   NV[float]  =  0.0
        init_seq7_value2:   NV[float]  =  0.0
        init_seq8_action:   NV[int]    =  0
        init_seq8_value1:   NV[float]  =  0.0
        init_seq8_value2:   NV[float]  =  0.0
        init_seq9_action:   NV[int]    =  0
        init_seq9_value1:   NV[float]  =  0.0
        init_seq9_value2:   NV[float]  =  0.0
        init_seq10_action:  NV[int]    =  0
        init_seq10_value1:  NV[float]  =  0.0
        init_seq10_value2:  NV[float]  =  0.0




from pydantic import BaseModel, validator
from typing import List
from enum import Enum 



##### ############
# Sequence
class INITSEQ(int, Enum):
    END = 0
    FIND_INDEX = 1
    FIND_REF_LE = 2
    FIND_REF_UE = 3
    FIND_LHW = 4
    FIND_UHW = 5

    DELAY = 6
    MOVE_ABS = 7
    MOVE_REL = 8
    CALIB_ABS = 9
    CALIB_REL = 10
    CALIB_SWITCH = 11

for E,v1,v2 in [
    ( INITSEQ.END,  "", "" ),
    ( INITSEQ.FIND_INDEX, "Fast Vel", "Slow Vel" ),
    ( INITSEQ.FIND_REF_LE, "Fast Vel", "Slow Vel" ),
    ( INITSEQ.FIND_REF_UE, "Fast Vel", "Slow Vel" ),
    ( INITSEQ.FIND_LHW, "Fast Vel", "Slow Vel" ),
    ( INITSEQ.FIND_UHW, "Fast Vel", "Slow Vel" ),
    ( INITSEQ.DELAY, "Delay [ms]", "" ),
    ( INITSEQ.MOVE_ABS, "Vel", "Pos" ),
    ( INITSEQ.MOVE_REL, "Vel", "Pos" ),
    ( INITSEQ.CALIB_ABS, "Pos", "" ),
    ( INITSEQ.CALIB_REL, "Pos", "" ),
    ( INITSEQ.CALIB_SWITCH, "Pos", "" ),
]:
    setattr(E, "var1", v1)
    setattr(E, "var2", v2)
del E,v1,v2



# ################################        
    
class SeqStepConfig(BaseModel):
    """ Data  Model for step configuration """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    index: int =  0    
    value1: float = 0.0
    value2: float = 0.0
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~
    class Config:                
        validate_assignment = True
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Method to save back the configuration     
    def cfgdict(self):
        return self.dict(exclude={"index"})
# ################################

class InitialisationConfig(BaseModel):
    """ Config Model for the initialisation sequence """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    sequence : List[str] = []
    END          : SeqStepConfig = SeqStepConfig(index=0) 
    FIND_INDEX   : SeqStepConfig = SeqStepConfig(index=1)
    FIND_REF_LE  : SeqStepConfig = SeqStepConfig(index=2)
    FIND_REF_UE  : SeqStepConfig = SeqStepConfig(index=3)
    FIND_LHW     : SeqStepConfig = SeqStepConfig(index=4)
    FIND_UHW     : SeqStepConfig = SeqStepConfig(index=5)  
    DELAY        : SeqStepConfig = SeqStepConfig(index=6)
    MOVE_ABS     : SeqStepConfig = SeqStepConfig(index=7)
    MOVE_REL     : SeqStepConfig = SeqStepConfig(index=8)
    CALIB_ABS    : SeqStepConfig = SeqStepConfig(index=9)
    CALIB_REL    : SeqStepConfig = SeqStepConfig(index=10)
    CALIB_SWITCH : SeqStepConfig = SeqStepConfig(index=11)
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~
    class Config:                
        validate_assignment = True        
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Validator Functions
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @validator('END', 'FIND_INDEX', 'FIND_REF_LE', 'FIND_REF_UE', 'FIND_LHW', 'FIND_UHW', 
               'DELAY', 'MOVE_ABS', 'MOVE_REL', 'CALIB_ABS', 'CALIB_REL' , 'CALIB_SWITCH')
    def force_index(cls, v, field):
        """ need to write the index """        
        v.index = getattr(INITSEQ, field.name)
        return v

    @validator('sequence')
    def validate_initialisation(cls,sequence):   
        """ Validate the list of sequence """ 
        for s in sequence:
            try:
                cls.__fields__[s]
            except KeyError:
                raise ValueError(f'unknown sequence step named {s!r}')
        return sequence
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Method to save back the configuration     
    def cfgdict(self):
        d = {'sequence':self.sequence}
        for seq in self.sequence:
            d[seq] = getattr(self, seq).cfgdict()            
        return d
# ################################



def init_sequence_to_cfg(initialisation, INITSEQ=INITSEQ):
    """ from a config initialisation dict return a dictionary of key/value for .cfg interface """            
    
    
    # set the init sequence    
    cfg_dict = {} 
    
    init_dict = initialisation.dict(exclude_none=True, exclude_unset=True)
    if not "sequence" in init_dict:        
        return cfg_dict
    
    # reset all sequence variable
    for i in range(1,11):
        cfg_dict["init_seq{}_action".format(i)] = INITSEQ.END.value
        cfg_dict["init_seq{}_value1".format(i)] = 0.0
        cfg_dict["init_seq{}_value2".format(i)] = 0.0
        
    for stepnum, step_name in enumerate(initialisation.sequence, start=1):
        step = getattr(initialisation, step_name)
        cfg_dict["init_seq%d_action"%stepnum] = step.index
        cfg_dict["init_seq%d_value1"%stepnum] = step.value1
        cfg_dict["init_seq%d_value2"%stepnum] = step.value2    
    return cfg_dict

